


/*
 * MapReduce application implementing an
 * iterative algorithm
 * 
 * End of Grade Project for UOC
 * 
 * First semester 2015-2016
 *
 * Computer Architecture and Operating Systems
 *
 * Francisco J. Alvarez Goikoetxea
 * 
 */

// We are going to need some imports

import java.io.IOException;
import java.lang.String;
import java.nio.file.Files;
import java.util.Arrays;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.File;
import java.io.FileWriter;
import java.io.BufferedWriter;



// The class Mapper, of course
import org.apache.hadoop.mapreduce.Mapper;
// The class Reducer
import org.apache.hadoop.mapreduce.Reducer;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.output.FileOutputFormat;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.FloatWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.io.ArrayWritable;


public class HadoopIter {
	
	
    /* We need a temporary directory for I/O */
    static private final Path TMP_DIR = new Path("TMP_Iteration");
	

    
 /**
 *
 * The Map class for the MapReduce iterative algorithm implementation.
 * 
 * Intended for Hadoop 1.2.1
 * 
 * The application implements the Jacobi method to solve systems of equations.
 * Several iterations are performed until a fixed accuracy is achieved.
 * 
 * 
 * @version 1.0, 28 Oct 2015
 * 
 * @author Francisco J. Alvarez Goikoetxea
 */
public static class IterMapper 
    extends Mapper<Object, Text, Text, Text> {
    


	
	
	    
    
    /*
        We must remember that a map() method is needed in the Mapper.
        This method is called once for each input (key,value) pair.
        
    */
    
    /**
     * map() method
     * 
     * As long as we are going to work with Hadoop primitives, we need to parse Text
     * to Float. We need to iterate through the input to convert the data from Text
     * to FloatWritable array. If we did this in the mapper, we would need to iterate
     * again in the reducer to convert data from the FloatWritable array to a more handy
     * data structure because the Hadoop arrays have just a few methods to work with them.
     * Instead iterating twice, which would not be very performance oriented, the mapper
     * will just pass the row number and the line of text to the reducer, where all the
     * processing will be performed. 
     * 
     * @param key is the the of the input (key,value) pair.
     * @param value is the value of the input (key,value) pair.
     * @param context
     * 
     * @throws IOException
     * @throws InterruptedException
     * 
     */
    public void map (Object key, Text value, Context context)
            throws IOException, InterruptedException {
        
        /* First, we convert the input value to String */	
    	String line = value.toString();
    	/* We split the line in two parts, the row number and the data line to pass
    	   to reducer */
    	String [] components = line.split(" ", 2);
  
    	
    	/* Path for the guess vector */
    	Configuration conf =new Configuration();
    	FileSystem fs = FileSystem.get(conf);
    	Path guessFile = new Path("guess");
    	if (!fs.exists(guessFile))
    	{
    		throw new IOException("guess file does not exist.");
    	}
    	
    	/* Opening the guess vector */
    	BufferedReader guessBuffer = new BufferedReader (new InputStreamReader (fs.open(guessFile)));
    	/* Getting the number of variables */
    	Integer unknowns = Integer.parseInt(guessBuffer.readLine().toString());
    	/* Unknowns vector as String[] */
    	String [] unknownsVector = new String[unknowns];
    	
    	
    	/* Reading all the variables and creating the unknowns vector */
    	for (int m = 0; m < unknowns; m++)
    	{
    		String [] lineArray = guessBuffer.readLine().toString().split(" ");
    		Integer lineNumber = Integer.parseInt(lineArray[0]);
    		unknownsVector[lineNumber-1] = lineArray[1]; 
    	}    	

     	
    	Text text1 = new Text();
    	Text text2 = new Text();
    	text1.set(components[0]);
    	text2.set(components[1] + " " + Arrays.deepToString(unknownsVector).substring(1, Arrays.deepToString(unknownsVector).length()-1).replaceAll(",", ""));
    	
    	/* Writing the output (key,value) pair */
        context.write(text1, text2);
        
       
    } // End method map()
    
    
    /*
     * 
     * Overriding the cleanup() method for it to delete the guess file
     * after each task is completed.
     * The reduce method will generate the guess file for the next iteration.
     * 
     */
    
    @Override
    protected void cleanup(Context context) throws IOException
    {    	
    	Configuration conf = new Configuration();
    	FileSystem fs = FileSystem.get(conf);
    	Path guessFile = new Path("guess");
    	fs.delete(guessFile, true);    	
    }
 
     
    
} // End class IterMapper



/**
 *
 * The Reduce class for the MapReduce log file processing application.
 * 
 * Intended for Hadoop 1.2.1
 * 
 * @version 1.0, 28 Oct 2015
 * 
 * @author Francisco J. Alvarez Goikoetxea
 */
public static class IterReducer
    extends Reducer<Text, Text, Text, Text> {
    
   
    /*
     * 
     * Overriding the setup method to write
     * the number of equations at the begining of the
     * guess file.
     * 
     */
	@Override
	protected void setup(Context context) throws IOException
	{
    	File newGuess = new File("guess");
    	FileWriter fw = new FileWriter(newGuess.getAbsoluteFile(), true);
    	BufferedWriter buffer = new BufferedWriter(fw);
    	
    	/* ESCRIBIENDO NUMERO DE INCOGNITAS */
    	/* BUSCAR ALTERNATIVA */
    	buffer.write("3");
    	buffer.newLine();
    	buffer.close();
	}
	
	
	/*
        There must be a reduce() method.
        This method is called once for each input (key, value) pair.    
    */
   
    
    /**
     * reduce() method
     * 
     * @param key
     * @param values
     * @param context
     * 
     * @throws IOException
     * @throws InterruptedException
     */
    public void reduce(Text key, Iterable<Text> value, Context context)
             throws IOException, InterruptedException{
    	
    	/* each unknown of the equation */
    	Text unknown = new Text();
    	/* estimated value of the unknown*/
    	Float estimatedValue = new Float(0.0f);
    	/* Text to String */
    	String [] floatNumbersString;
    	/* number of unknowns */
    	Integer numberOfUnknowns;
    	
    	
    	/* auxiliary variable */
    	String aux = "";
    	
    	/* naming the unknowns as x1, x2, x3, ... */
    	unknown.set(key.toString());
    	
    	/* Row of the equation */
    	Integer row = 0;
    	row = Integer.parseInt(key.toString());
    	
    	/* This should be not necessary, but MapReduce  obligates us to */
    	/* As long as the value received by  the reducer is an Iterable
    	 	we have to iterate, despite of having just one value */
    	for (Text val : value){          
        aux = aux + val.toString();
        }
    	
      	
    	/* We split the string in the corresponding float values */
    	floatNumbersString = aux.split(" ");    	
    	  	
    	/* How many unknowns? */
    	numberOfUnknowns = (floatNumbersString.length-1)/2;
    	
    	/* CoefficientsVector */
    	Float [] coefficientsVector = new Float[numberOfUnknowns];
    	for (Integer i = 0; i < numberOfUnknowns; i++)
    	{
    		coefficientsVector[i] = Float.parseFloat(floatNumbersString[i]);
      	}
    	
    	/* value of the equation */
    	Float equals = new Float(0.0f);
    	equals = Float.parseFloat(floatNumbersString[numberOfUnknowns]);
    	
    	/* Vector of estimated values */
    	Float [] vectorOfEstimatedValues = new Float[numberOfUnknowns];
    	for (Integer j = (floatNumbersString.length - (numberOfUnknowns)); j < floatNumbersString.length;j++)
    	{    		
    		vectorOfEstimatedValues[j - (numberOfUnknowns + 1)] = Float.parseFloat(floatNumbersString[j]);
    	}
    	
    	
    	/* We start calculating the estimated value of xi */
    	Float aii = new Float(coefficientsVector[(row - 1)]);
    	
    	Float sum = new Float(0.0f);
    	
    	for (Integer k = 0; k < numberOfUnknowns; k++)
    	{    		
    		if (k != (row-1))
    			sum = sum + (coefficientsVector[k]*vectorOfEstimatedValues[k]);    		
    	} 	
    	
    	estimatedValue = (equals - sum) / aii;
        
    	/* from Float to FloatWritable */
     	String salida = Arrays.deepToString(coefficientsVector).substring(1, Arrays.deepToString(coefficientsVector).length()-1).replaceAll(",", "") + " " + equals.toString();
    	Text result = new Text();
    	result.set(salida);
    	
    	/* We need to write to the guess file to comunicate with the next job */
    	Configuration conf =new Configuration();
    	FileSystem fs = FileSystem.get(conf);
    	//Path guessFile = new Path("guess");

    	/* The previous guess file has been deleted by mapper */
    	File newGuess = new File("guess");
    	FileWriter fw = new FileWriter(newGuess.getAbsoluteFile(), true);
    	BufferedWriter buffer = new BufferedWriter(fw);
    	
    	buffer.write(unknown.toString() + " " + estimatedValue.toString());
    	buffer.newLine();
    	buffer.close();
    	
    	context.write(unknown, result);
        
        
    } // End method reduce()    
    
} // End class IterReducer



/**
 *
 * The Driver for the MapReduce log file processing application
 * as the main() method.
 * 
 * Intended for Hadoop 1.2.1
 * 
 * @version 2.0, 02 Nov 2015
 * 
 * @author Francisco J. Alvarez Goikoetxea
 *
 * @param args - arguments for main()
 * 
 * @throws Exception
 * 
 */
    

    public static void main(String [] args) throws Exception
    {
    	
    	/* Credits */
    	System.out.println("\nHadoopIter");
    	System.out.println("Francisco J. Alvarez Goikoetxea");
    	System.out.println("\nVersion: 2.0");
    	System.out.println("\nCompile: 215\n\n");
    	
    	/* Number of iterations */
    	Integer iterations = Integer.parseInt(args[1]);
    
        /* First we set a handle for the Configuration instance for the job */
        Configuration conf = new Configuration();
        /* Creating the job */
        Job job0 = Job.getInstance(conf, "HadoopIter0");
        /* Setting the encapsulating JAR class */
        job0.setJarByClass(HadoopIter.class);
        /* Setting the mapper class */
        job0.setMapperClass(IterMapper.class);
        /* Setting the reducer class */
        job0.setReducerClass(IterReducer.class);
        /* 
        As long as the types of the output pair of the map class differ
        from the input ones, we must tell MapReduce what they are
        */
        job0.setMapOutputKeyClass(Text.class);
        job0.setMapOutputValueClass(Text.class);
        
        
        
        /* We create a connection to the file system */
        final FileSystem fs = FileSystem.get(conf);
        
        if (fs.exists(TMP_DIR))
        {        	
        	throw new IOException("Temporary directory already exists. Remove it first.");
        }

        
        /* First input directory */
        FileInputFormat.addInputPath(job0, new Path(args[0]));
        /* First output directory */
        Path outDir0 = new Path(TMP_DIR, "out_iter_0");
        FileOutputFormat.setOutputPath(job0, outDir0);
        /* First job execution */
        job0.waitForCompletion(true);
        
        
        /* Iteratin' */
        
        for (Integer n = 1; n < iterations; n++)
        {
        	
        /* First we set a handle for the Configuration instance for the job */
        Configuration config = new Configuration();
        /* Creating the job */
        Job job = Job.getInstance(config, "HadoopIter" + n);
        /* Setting the encapsulating JAR class */
        job.setJarByClass(HadoopIter.class);        
        /* Setting the mapper class */
        job.setMapperClass(IterMapper.class);
        /* Setting the reducer class */
        job.setReducerClass(IterReducer.class);
        
        job.setMapOutputKeyClass(Text.class);
        job.setMapOutputValueClass(Text.class);
        
        	
        FileInputFormat.addInputPath(job, new Path(TMP_DIR, "out_iter_" + (n-1)));

 
        Path outDir = new Path(TMP_DIR, "out_iter_" + n);
        FileOutputFormat.setOutputPath(job, outDir);       
        
        /* Finally, we wait for job completion. */
        job.waitForCompletion(true);
        
        } // End of iterations loop
        
        /* Deleting temporary directory */
        fs.delete(TMP_DIR, true);
        
        
    } // End main() - The end of the Driver
    
 

    
} // End class HadoopIter
