/*
 * MapReduce application for log processing
 * 
 * End of Grade Project for UOC
 * 
 * First semester 2015-2016
 *
 * Computer Architecture and Operating Systems
 *
 * Francisco J. Alvarez Goikoetxea
 * 
 */

// We are going to need some imports

// Exceptions for error management
import java.io.IOException;

import java.util.StringTokenizer;

// The class Mapper, of course
import org.apache.hadoop.mapreduce.Mapper;
// The class Reducer
import org.apache.hadoop.mapreduce.Reducer;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.output.FileOutputFormat;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.Text;


public class HadoopLog {
    
 /**
 *
 * The Map class for the MapReduce log file processing application.
 * 
 * Intended for Hadoop 1.2.1
 * 
 * The LogMapper reads a text log file.
 * As input takes a <Text, Text> pair.
 * As output provides a <Object, IntWritable> pair.
 * 
 * 
 * @version 1.0, 03 Oct 2015
 * 
 * @author Francisco J. Alvarez Goikoetxea
 */
public static class LogMapper 
    extends Mapper<Object, Text, Text, IntWritable> {
    
    /* Here they come the variable definitions */
    /*
        As long as they are private, there is no need for getters and setters.
    */
    
    /* We need the (key,value) output pair */
    /* The key is going to be the url that accesed the NASA site */
    private Text loggedUrl = new Text();
    /* The value is going to be the number of bytes downloaded */
    private IntWritable bytesDownloaded = new IntWritable();
    /* We need a intermediate Integer value to store the bytes */
    private Integer number;
    
    /*
        We must remember that a map() method is needed in the Mapper.
        This method is called once for each input (key,value) pair.
    */
    
    /**
     * map() method
     * 
     * @param key is the the of the input (key,value) pair.
     * @param value is the value of the input (key,value) pair.
     * @param context
     * 
     * @throws IOException
     * @throws InterruptedException
     * 
     */
    public void map (Object key, Text value, Context context)
            throws IOException, InterruptedException {
        
        /* We need to split the line to get the data we need */
        
        /* First, we convert the line to String */
        String line = value.toString();
        
        /* Now, we split the line using spaces as separators */
        String [] lineParts = line.split(" ");
        
        /* The key is going to be the IP or the URL (lineParts[0]) */
        loggedUrl.set(lineParts[0]);
        /* The value is going to be the number of downloaded bytes (last field of the array) */
        /* Sometimes we have a "-" instead a number. We set the value to 0 */
        if(lineParts[(lineParts.length)-1].equals("-")) {
	  bytesDownloaded.set(0);
	  }
	else {
	  System.out.println("Linea: " + line);
	  System.out.println("Ultimo campo: " + lineParts[(lineParts.length)-1]);
	  bytesDownloaded.set(number.parseInt(lineParts[(lineParts.length)-1]));
	  }
        
        /* Writing the output (key,value) pair */
        context.write(loggedUrl,bytesDownloaded);
        
        
        
    } // End method map()    
    
} // End class LogMapper



/**
 *
 * The Reduce class for the MapReduce log file processing application.
 * 
 * Intended for Hadoop 1.2.1
 * 
 * @version 1.0, 03 Oct 2015
 * 
 * @author Francisco J. Alvarez Goikoetxea
 */
public static class LogReducer
    extends Reducer<Text, IntWritable, Text, IntWritable> {
    
    /* We need a IntWritable to store the final result */
    /* As long as it is private, there is no need for getter and setter */
    private IntWritable result = new IntWritable();
    
    
    /*
        There must be a reduce() method.
        This method is called once for each input (key, value) pair.    
    */
    
    /**
     * reduce() method
     * 
     * @param key
     * @param values
     * @param context
     * 
     * @throws IOException
     * @throws InterruptedException
     */
    public void reduce(Text key, Iterable<IntWritable> values, Context context)
            throws IOException, InterruptedException {
        
        /* Some variables needed */
        /* Maximun bytes downloaded */
        int max = 0;
        /* A temporary value to store current values */
        int temporary = 0;        
        
        for (IntWritable val : values) {
            
            /* Getting the input value */
            temporary = val.get();
            /* If the new value is greater than max, it is set as max */
            if (temporary > max)
                max = temporary;        
        
        }
        
        /* Setting the result value */
        result.set(max);
        
        /* writing the output (key,value) pair */
        context.write(key,result);
        
    } // End method reduce()    
    
} // End class LogReducer



/**
 *
 * The Driver for the MapReduce log file processing application
 * as the main() method.
 * 
 * Intended for Hadoop 1.2.1
 * 
 * @version 1.0, 03 Oct 2015
 * 
 * @author Francisco J. Alvarez Goikoetxea
 *
 * @param args - arguments for main()
 * 
 * @throws Exception
 * 
 */
    

    public static void main(String [] args) throws Exception
    {
    
          
	/* Credits */
	System.out.println("\nHadoopLog");
	System.out.println("Francisco J. Alvarez Goikoetxea");
	System.out.println("\nVersion: 1.2");
	System.out.println("\nCompile: 26\n\n");
    
    
    
        /* First we set a handle for the Configuration instance for the job */
        Configuration conf = new Configuration();
        /* Creating the job */
        Job job = Job.getInstance(conf, "hadoop_log");
        /* Setting the encapsulating JAR class */
        job.setJarByClass(HadoopLog.class);
        /* Setting the mapper class */
        job.setMapperClass(LogMapper.class);
        /* Setting the reducer class */
        job.setReducerClass(LogReducer.class);
        /* 
        As long as the types of the output pair of the map class differ
        from the input ones, we must tell MapReduce what they are
        */
        job.setMapOutputKeyClass(Text.class);
        job.setMapOutputValueClass(IntWritable.class);
        
        /*
        We must set the input and output paths to the input and output files.
        We set them from the arguments.
        */
        FileInputFormat.addInputPath(job, new Path(args[0]));
        FileOutputFormat.setOutputPath(job, new Path(args[1]));
        
        /* Finally, we wait for job completion. */
        System.exit(job.waitForCompletion(true)? 0 : 1);
        
        
    } // End main() - The end of the Driver
    
 

    
} // End class HadoopLog
