/*
 * MapReduce traffic generator & analyzer
 * 
 * End of Grade Project for UOC
 * 
 * First semester 2015-2016
 *
 * Computer Architecture and Operating Systems
 *
 * Francisco J. Alvarez Goikoetxea
 * 
 * 17th of December, 2015
 * 
 * Version - v0.4
 * 
 */



import java.io.*;
import java.net.URL;

import javax.json.*;


public class MRGen {
	
	public static void main (String [] args) throws Exception {
		
		
		/* Some credits */
		System.out.println("\n**********************************************\n");
		System.out.println("\n\tMapReduce job Launcher\n");
		System.out.println("\tFrancisco J. Alvarez Goikoetxea\n");
		System.out.println("\tVersion 1.0\n");
		System.out.println("\tTFG - UOC");
		System.out.println("\n**********************************************\n");
		
		
		/* First some argument control */
		/*
		 * 
		 * args[0] -> input directory generic name
		 * args[1] -> output directory generic name
		 * args[2] -> number of jobs the generator is going to send
		 * args[3] -> number of iterations to be performed by the job
		 * 
		 * Initially args[3] is going to be an Integer. It should become an
		 * Integer[], with a different number of iterations per job. 
		 * 
		 */
		if (args.length < 4)
		{
			System.out.println("\nNumber of parameters not correct.");
			System.out.println("Parameters: /input_directory /output_directory number_of_executions number_of_iterations\n");
			System.exit(-2);
		}
		
		/* Setting the number of executions */
		Integer executions = new Integer(Integer.parseInt(args[2]));
		
		/* Executions loop */
		for (int e = 0; e < executions; e++)
		{
		
			System.out.println("\nTest execution number: " + e + "\n");
		
		try
		{
			
		/* The input directory set from the arguments and from the execution number */
		String inputDirectory = args[0] + e;
		/* The output directory set from the arguments and from the execution number */
		String outputDirectory = args[1] + e;
		Integer iterations = new Integer(Integer.parseInt(args[3]));
		
		/* We need to get the name of the jar file inside the input directory */
		/* The path to the input directories */
		String currentJarDirectory = "input" + e;
		File folder = new File("/home/goiko/programas/hadoop-2.7.1/" + currentJarDirectory);
		/* The list of files inside the previous path (there should be only one) */
		File[] listOfFiles = folder.listFiles();
		String jarFileName = listOfFiles[0].getName();
		
		String currentJarPath = "./" + currentJarDirectory + "/" + jarFileName;
		
			
		/* The String with the command line */
		/* As long as ProcessBuilder does not parse the command line, it has to be a String array */
		String [] arguments = new String[]{"bin/hadoop", "jar", inputDirectory + "/" + jarFileName, "HadoopIter", "/user/goiko/input0", iterations.toString()};
		
		
		/* ProcessBuilder object */
		ProcessBuilder pb = new ProcessBuilder(arguments);	
		
		/* Redirecting output and error streams */
		/* As long as the buffers for both are limited, it is needed to drain them quickly */
		/* for the external process not to block */
		File output = new File("./output.txt");
		File err = new File("./errors.txt");
		pb.redirectOutput(output);
		pb.redirectError(err);
		
		/* Starting the external process */
		Process myApp = pb.start();
		
		/* Getting the exit value to know what's happened */
		int exitVal = myApp.waitFor();
		System.out.println("\nFinish code: " + exitVal + "\n");
		} // End try
		catch (Exception excpt)
		{			
			excpt.printStackTrace();
			System.exit(-1);
		} // End catch
		
		System.out.println("\n\n\t\tTest execution " + e + " successful\n\n");
		
		
	    /* FROM HERE ON -> GETTING DATA FROM HISTORY SERVER */
        /* First, we need to get the last app ID. As with Spark, we need a little trick */
        /* Reading the last application run from the history server */
        
        /* First, we need to wait for the server to get the last job */
        /* Ten seconds to make sure. Seven proved not to be enough. */
        System.out.println("\n\n**************************************************************************************\n"
        		+ "\tWaitin' for the History Server to get the last job.\n"
        		+ "\tBe patient dude. This is a test, we're not goin' for performance here."
        		+ "\n**************************************************************************************\n\n");
        Thread.sleep(10000);
        
        /* First thing, we need the job ID */
        
        URL histServer = new URL("http://localhost:19888/ws/v1/history/mapreduce/jobs");
        InputStream histStream = histServer.openStream();
        
        /* Getting the full JSON file */
        JsonReader readJson = Json.createReader(histStream);
        JsonObject completeObject = readJson.readObject();
        /* Getting the "jobs" JSON object */
        JsonObject jobsObj = completeObject.getJsonObject("jobs");
        /* Getting the "job" object */
        /* An Array containing all the jobs executed - among other properties, the ID of every job executed */
        JsonArray jobObj = jobsObj.getJsonArray("job");
        /* The job we are interested in is the last one */
        /* Getting the jobs array size */
        Integer size = jobObj.size();
        /* Readin' the last object */
        JsonObject lastJobProperties = jobObj.getJsonObject(size - 1);
        /* Getting the ID */
        JsonValue lastJobID = lastJobProperties.get("id");
        
        String jobid = lastJobID.toString().substring(1, lastJobID.toString().length() - 1);
        
        
		System.out.println("\nLast job ID: " + jobid + "\n");
		
		/* Getting the metrics from the history server */
		/* They are to be written inside a XML file for processing */
		try
		{
			/* The history server address */
			/* Getting in JSON format */
			URL histServ = new URL("http://localhost:19888/ws/v1/history/mapreduce/jobs/" + jobid + "/counters");
			
			/* We need to load the JSON file into a JSON Object to extract data from it */

			/* Opening the file in the history server */
			InputStream in = histServ.openStream();
			/* Creating the JSON object contained in the file */
			JsonReader rdr = Json.createReader(in);
			JsonObject obj = rdr.readObject();
				
			/* The structure of the file is quite complex, with nested */
			/* objects and arrays. We have to select the correct values from it. */
			JsonObject jobCounters = obj.getJsonObject("jobCounters");
			JsonArray counterGroup = jobCounters.getJsonArray("counterGroup");
			JsonObject object3 = counterGroup.getJsonObject(2);
			JsonArray counter = object3.getJsonArray("counter");
			JsonObject cpuTimeObject = counter.getJsonObject(16);
			JsonObject physicalMemoryObject = counter.getJsonObject(17);
				
			JsonValue cpu_time = cpuTimeObject.get("totalCounterValue");
			JsonValue physical_memory = physicalMemoryObject.get("totalCounterValue");
				
			/* Now getting the bytes read from disk */
			JsonObject object4 = counterGroup.getJsonObject(4);
			JsonArray counter2 = object4.getJsonArray("counter");
			JsonObject totalBytesRead = counter2.getJsonObject(0);
			JsonValue bytesRead = totalBytesRead.get("totalCounterValue");

			/* And now going for the bytes written to disk */
			JsonObject object5 = counterGroup.getJsonObject(5);
			JsonArray counter3 = object5.getJsonArray("counter");
			JsonObject totalBytesWritten = counter3.getJsonObject(0);
			JsonValue bytesWritten = totalBytesWritten.get("totalCounterValue");
		
			/* Appending to metrics.txt (if it does not exists, it creates it */
			BufferedWriter wrToFile = new BufferedWriter(new FileWriter("metrics.txt", true));
			wrToFile.write("\n\nTest Execution: " + e);
			wrToFile.write("\nJob ID: " + jobid + "\n");
			wrToFile.write("\nCPU time (ms): " + cpu_time);
			wrToFile.write("\nTotal Phys. Memory (MB): " + (Long.parseLong(physical_memory.toString())/1024)/1024);
			wrToFile.write("\nTotal KB read from disk: " + (Float.parseFloat(bytesRead.toString())/1024));
			wrToFile.write("\nTotal KB written to disk: " + (Float.parseFloat(bytesWritten.toString())/1024));
			
			/* Getting the total elapsed time */
			/* We have to get the JSON object from a different file */
			//URL histServer2 = new URL("http://localhost:19888/ws/v1/cluster/app/application" + jobid.substring(3));
			URL histServer2 = new URL("http://localhost:19888/ws/v1/history/mapreduce/jobs/" + jobid);
			/* Opening the file in the history server */
			InputStream in2 = histServer2.openStream();
			/* Creating the JSON object contained in the file */
			JsonReader rdr2 = Json.createReader(in2);
			JsonObject obj2 = rdr2.readObject();
			
			/* Getting the time values */
			JsonObject jsonJob = obj2.getJsonObject("job");
			/* Finish time */
			JsonValue finishTime = jsonJob.get("finishTime");
			Long ft = Long.parseLong(finishTime.toString());
			/* Submit time */
			JsonValue submitTime = jsonJob.get("submitTime");
			Long st = Long.parseLong(submitTime.toString());
			/* Finish time */
			JsonValue startTime = jsonJob.get("startTime");
			Long stt = Long.parseLong(startTime.toString());
			
			/* Logging the times */
			wrToFile.write("\nTotal elapsed time: " + (ft-st)/1000 + "seconds");
			wrToFile.write("\nTotal execution time: " + (ft-stt)/1000 + "seconds");
			
			
			/* Closing the file */
			wrToFile.close();
			
			
			
		} // End try
		
		catch (Exception excpt)
		{
			excpt.printStackTrace();
		} // End catch
        		
		
		} // END OF THE EXECUTIONS LOOP
		
		
	} // End main() method

} // End class MRGen


