/*
 * Spark application to implement a iterative
 * numeric method.
 * Jacobi method for n equations in n unknowns
 * 
 * End of Grade Project for UOC
 * 
 * First semester 2015-2016
 *
 * Computer Architecture and Operating Systems
 *
 * Francisco J. Alvarez Goikoetxea
 * 
 */

/* We are going to need some imports */

/* From Java itself */

import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Map;
import java.io.IOException;

/* And, of course, from Spark */

import org.apache.spark.api.java.JavaRDD;
import org.apache.spark.api.java.JavaPairRDD;
import org.apache.spark.api.java.JavaSparkContext;
import org.apache.spark.api.java.function.DoubleFunction;
import org.apache.spark.api.java.function.Function;
import org.apache.spark.api.java.function.Function2;
import org.apache.spark.api.java.function.PairFunction;
import scala.Tuple2;
import org.apache.spark.SparkConf;
import org.apache.spark.SparkFiles;



class Compute implements Function<String, String>
{        	
	
	private Float [] estimatedValues;
	
	/* Number of variables in the system of equations */
	private Integer variable_number = 3;
	
	/* Constructor with the Float array of estimated values as argument */
	Compute (Float [] estimatedValues) { this.estimatedValues = estimatedValues; }
	
	public String call (String s) {
		
		
		String result = new String();
		
		/* The String "s" has all the information needed so far */
		/* We are to split it in the elements needed to estimate the new guess */
		String [] wholeString = s.split(" ");
		
		/* Array of Floats representing the equation coefficients corresponding to "key" row */
		Float [] coefficients = new Float [variable_number];
		for (int i = 0; i < variable_number; i++)
		{
			coefficients[i] = Float.parseFloat(wholeString[i]);
		}
		
		/* The vector B of the system */
		Float [] equals = new Float[variable_number];
		for (int i = 0; i < variable_number; i++){
			equals[i] = Float.parseFloat(wholeString[i+variable_number]);
		}
		
		
		Integer row = Integer.parseInt(wholeString[2*variable_number]);
	
		
		/* Calculating the new estimated value */
		Float sum = new Float(0.0f);
		for (int i = 0; i < variable_number; i++)
		{
			if (i != (row-1))
				sum = sum + coefficients[i] * estimatedValues[i];
		}
		
		Float guess = new Float(0.0f);
		guess = (equals[row-1] - sum) / coefficients[row-1];
		
		
		result = guess.toString();
		
		return result;
		
	}
	
}





public class SparkIter {
	
	
	/**
	 *
	 * The main method for the Spark log file processing application.
	 * 
	 * Intended for Spark 1.4.1
	 * 
	 * @version 1.4, 02 Dec 2015
	 * 
	 * @author Francisco J. Alvarez Goikoetxea
	 *
	 * @param args - arguments for main()
	 * 
	 * @throws Exception
	 * 
	 */
	

	public static void main (String [] args) throws Exception
	{
		
		/* Context object to communicate with the Spark framework */
		final JavaSparkContext sc = new JavaSparkContext ();
		
		/* Mutable array to contain the initial guess */
		final Float [] estimatedValues = new Float[] {0.0f, 0.0f, 0.0f};
		
		final Integer variable_number = 3;
		
		/* We need an argument to specify the number of iterations */
		Integer iterations;
		if (args.length == 0)
		{
			System.out.println("\nYou must specify the number of iterations to be performed.\n");
			System.exit(3);
		}
		iterations = Integer.parseInt(args[0]);

		
		/* Loading the system of equations from file */
		/* We create the RDD to work on the file on parallel */
        JavaRDD<String> lines = sc.textFile("/home/goiko/datos/system.txt");
        
        /* We need another RDD to load the initial guess */
        //JavaRDD<String> guess = sc.textFile("/home/goiko/datos/guess");
        
        /* Credits. */
    	System.out.println("\nSparkIter v1.4");
    	System.out.println("Francisco J. Alvarez Goikoetxea");
    	System.out.println("\nVersion: 1.2");
    	System.out.println("\nCompile: 153\n\n");
        
        
        /* We map the lines of the data file to a (key, value) pair */
        /* We use the mapToPair() method to create a JavaPairRDD */
        /* The PairFuntion needs three arguments:
         	- String -> The line from the text file as input
         	- Integer -> The row of the coefficient matrix (key of the output pair)
         	- String -> The value of the output pair (is going to be the same as the input) */
        
        JavaPairRDD<Integer, String> mappingToPair = lines.mapToPair(new PairFunction<String, Integer, String>()
        		{
        	
        		public Tuple2<Integer, String> call (String s) {
        			
        			
        			Integer key;
        			
        			String[] tokens = s.split(" ", 2);
        			key = Integer.valueOf(tokens[0]);
        			String coefficients = tokens[1];
        			
        			Tuple2<Integer, String> mapResult = new Tuple2<Integer, String>(key, coefficients);
        			
       			
        			return mapResult;
        			
        		}
        	
        		});
        


        
        
        
       

        /* Mappings change the values of a PairRDD, while keeping the key, as in the previous RDD */
        /* Now we need to map the values to a new value including the new guess instead of the old one */
        
        /* The resultingPairRDD is going to have the same (K, V) types than the previous one */
        String estimatedString = new String();
        String [] estimatedArray = new String[variable_number];
        
        for (int n = 0; n < iterations; n++)
        {
        
        JavaPairRDD<Integer, String> guessingPair = mappingToPair.mapValues(new Compute (estimatedValues));
        
		
        String aux = guessingPair.values().collect().toString();
        estimatedString = aux.substring(1, aux.length()-1).replace(", ", " ");
        estimatedArray = estimatedString.split(" ");
        System.out.println("\n");
        for (int i = 0; i < variable_number; i++)
        {
        	estimatedValues[i] = Float.parseFloat(estimatedArray[i]);
        	System.out.println("Variable" + i + " = " + estimatedArray[i]);
        }
        System.out.println("\n");
        
        
        } // end of the iterations loop
        
        System.out.println("\n");
        
        
        //guessingPair.sortByKey().repartition(1).saveAsTextFile("/home/goiko/datos/output.txt");
        //reducerInput.sortByKey().repartition(1);

        
        
        /* Stopping the SparkContext */
        sc.stop();
        
	} // End method main()

} // End class SparkIter
