/*
 * Spark application for log processing
 * 
 * End of Grade Project for UOC
 * 
 * First semester 2015-2016
 *
 * Computer Architecture and Operating Systems
 *
 * Francisco J. Alvarez Goikoetxea
 * 
 */

/* We are going to need some imports */

/* From Java itself */

import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.io.IOException;

/* And, of course, from Spark */

import org.apache.spark.api.java.JavaRDD;
import org.apache.spark.api.java.JavaPairRDD;
import org.apache.spark.api.java.JavaSparkContext;
import org.apache.spark.api.java.function.DoubleFunction;
import org.apache.spark.api.java.function.Function;
import org.apache.spark.api.java.function.Function2;
import org.apache.spark.api.java.function.PairFunction;
import scala.Tuple2;
import org.apache.spark.SparkConf;
import org.apache.spark.SparkFiles;


public class SparkLog {
	
	
	/**
	 *
	 * The main method for the Spark log file processing application.
	 * 
	 * Intended for Spark 1.4.1
	 * 
	 * @version 1.0, 14 Oct 2015
	 * 
	 * @author Francisco J. Alvarez Goikoetxea
	 *
	 * @param args - arguments for main()
	 * 
	 * @throws Exception
	 * 
	 */
	

	public static void main (String [] args) throws Exception
	{
		
		/* First thing we need is a Context to write to */
		final JavaSparkContext sc = new JavaSparkContext ("spark://goiko:7077", "SparkLog");
		
		/* Loading the data from text file */
		/* We create the RDD to work on the file on parallel */
        JavaRDD<String> lines = sc.textFile("/home/goiko/datos/log.txt");
        
        /* As long as we are using JDK 1.7, we cannot use lambda expressions.
         	We will implement the Function interface to create anonymous classes instead*/
        
        /* We map the lines of the data file to a (key, value) pair */
        /* We use the mapToPair() method to create a JavaPairRDD */
        /* The PairFuntion needs three arguments:
         	- String -> The line from the text file as input
         	- String -> The Key of the output pair (the URL)
         	- Integer -> The value of the output pair (the bytes downloaded)*/
        
        JavaPairRDD<String, Integer> reducerInput = lines.mapToPair(new PairFunction<String, String, Integer>()
        		{
        	
        		public Tuple2<String, Integer> call (String s) {
        			
        			/* Inside the anonymous class itself 
        			 	what's to be done is to be done here */
        			
        			Integer bytes;
        			
        			String[] tokens = s.split(" ");
        			// System.out.println("URL: " + tokens[0] + "\tBytes: " + tokens[tokens.length-1]);
        			String url = tokens[0];
        			if (tokens[tokens.length-1].equals("-"))
        			{bytes = 0;}
        			else 
        			{bytes = Integer.valueOf(tokens[tokens.length-1]);}
        			Tuple2<String, Integer> mapResult = new Tuple2<String, Integer>(url, bytes);
        			
        			return mapResult;
        			
        		}
        	
        		});
        		
        
        /* We will use the reduceByKey() method to implement the "reducer"  */
        /* A funtion to get the maximum value of downloaded bytes is needed to perform
         	the reduceByKey action */
         JavaPairRDD<String, Integer> reduceResult = reducerInput.reduceByKey(
        		
        		new Function2<Integer, Integer, Integer>()
        		{
        			
        			public Integer call(Integer i1, Integer i2)
        			{
        				
        				if (i1 > i2)
        					return i1;
        				else
        					return i2;     				
        				
        			}
        			
        		}
        		
        		);
        

        reduceResult.sortByKey().repartition(1).saveAsTextFile("/home/goiko/datos/output.txt");
        
        
        
        /* Stopping the SparkContext */
        sc.stop();
        
	} // End method main()

} // End class SparkLog
