/*
 * Spark application to generate traffic
 * for test performing
 * 
 * End of Grade Project for UOC
 * 
 * First semester 2015-2016
 *
 * Computer Architecture and Operating Systems
 *
 * Francisco J. Alvarez Goikoetxea
 * 
 * 12 Dic 2015
 * 
 */

/* We are going to need some imports */

/* From Java itself */

import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.io.IOException;
import java.util.Date;

/* And, of course, from Spark */

import org.apache.spark.api.java.JavaRDD;
import org.apache.spark.rdd.RDD;

import org.apache.spark.api.java.JavaPairRDD;
import org.apache.spark.api.java.JavaSparkContext;
import org.apache.spark.api.java.function.DoubleFunction;
import org.apache.spark.api.java.function.Function;
import org.apache.spark.api.java.function.Function2;
import org.apache.spark.api.java.function.PairFunction;
import scala.Tuple2;
import org.apache.spark.SparkConf;
import org.apache.spark.SparkFiles;

import java.io.*;
import java.net.URL;

import javax.json.*;




public class SparkTrafficGen {
	
	
	/**
	 *
	 * The main method for the Spark log file processing application.
	 * 
	 * Intended for Spark 1.4.1
	 * 
	 * @version 1.0, 14 Oct 2015
	 * 
	 * @author Francisco J. Alvarez Goikoetxea
	 *
	 * @param args - arguments for main()
	 * 
	 * @throws Exception
	 * 
	 */
	

	public static void main (String [] args) throws Exception
	{
		
		
		/* Some credits */
		System.out.println("\n\nSpark Traffic Generator");
		System.out.println("\nFrancisco J. Alvarez Goikoetxea");
		System.out.println("\nVersion: 1.0");
		System.out.println("Compile: 93\n\n");
		
		
		
		/* We need to check the number of arguments, that is:
		 * 
		 *  The input directories name: input*
		 *  The output directories name: output*
		 *  The number of executions: n
		 *  
		 *  If we have 3 executions, the inouts are to be taken from:
		 *  input0, input1 and input2
		 *  The outputs are going to be written to
		 *  output0, output1 and output2
		 *  
		 *  */
		
		
		if (args.length < 3)
		{
			System.out.println("\nNumber of arguments is wrong.");
			System.out.println("\nArguments must be: input_dir_name output_dir_name number_of_executions\n");
			System.exit(-5);
		}
		
		
		/* Integer containing the number of executions to be performed */
		Integer executions = new Integer(0);
		executions = Integer.parseInt(args[2]);
		

		/* STARTING THE EXECUTIONS LOOP */
		
		/* Now we create as many executions as indicated as a parameter */
		for (int e = 0; e < executions; e++)
		{
			
		/* We need a configuration object if we want to be able to get the appId */
		SparkConf conf = new SparkConf();
		conf.setMaster("spark://goiko:7077");
		conf.setAppName("Spark Traffic Generator");
		
		/* First thing we need is a Context to write to */
		/* This hasn't the applicationId() method, and I need the hell it */
		final JavaSparkContext sc = new JavaSparkContext (conf);
		/* Goin' for the bloody applicationId() method */
		//final SparkContext spc = new SparkContext(conf);		
		
		System.out.println("\nTest execution: " + e + "\n");		
								
		/* Loading the data from text file */
		/* We create the RDD to work on the file on parallel */
		String inputFile = "/home/goiko/datos/" + args[0] + e + "/log.txt";
		/* More changes because of the use of SparkContext instead of JavaSparkContext */
        JavaRDD<String> lines = sc.textFile(inputFile);
		//RDD<String> lines = sc.textFile(inputFile,1);
        
       
        /* As long as we are using JDK 1.7, we cannot use lambda expressions.
         	We will implement the Function interface to create anonymous classes instead*/
        
        /* We map the lines of the data file to a (key, value) pair */
        /* We use the mapToPair() method to create a JavaPairRDD */
        /* The PairFuntion needs three arguments:
         	- String -> The line from the text file as input
         	- String -> The Key of the output pair (the URL)
         	- Integer -> The value of the output pair (the bytes downloaded)*/
        
        JavaPairRDD<String, Integer> reducerInput = lines.mapToPair(new PairFunction<String, String, Integer>()
        		{
        	
        		public Tuple2<String, Integer> call (String s) {
        			
        			/* Inside the anonymous class itself 
        			 	what's to be done is to be done here */
        			
        			Integer bytes;
        			
        			String[] tokens = s.split(" ");
        			// System.out.println("URL: " + tokens[0] + "\tBytes: " + tokens[tokens.length-1]);
        			String url = tokens[0];
        			if (tokens[tokens.length-1].equals("-"))
        			{bytes = 0;}
        			else 
        			{bytes = Integer.valueOf(tokens[tokens.length-1]);}
        			Tuple2<String, Integer> mapResult = new Tuple2<String, Integer>(url, bytes);
        			
        			return mapResult;
        			
        		}
        	
        		});
        
                
        /* We will use the reduceByKey() method to implement the "reducer"  */
        /* A funtion to get the maximum value of downloaded bytes is needed to perform
         	the reduceByKey action */
         JavaPairRDD<String, Integer> reduceResult = reducerInput.reduceByKey(
        		
        		new Function2<Integer, Integer, Integer>()
        		{
        			
        			public Integer call(Integer i1, Integer i2)
        			{
       				
        				if (i1 > i2)
        					return i1;
        				else
        					return i2;     				
        				
        			}
        			
        		}
        		
        		);
        

         String outputDir = "/home/goiko/datos/" + args[1] + e;
         reduceResult.sortByKey().repartition(1).saveAsTextFile(outputDir);
         
      
        /* Stopping the JavaSparkContext */
        sc.stop();
        

        /* FROM HERE ON -> GETTING DATA FROM HISTORY SERVER */
        /* First, we need to get the last app ID. It ain't as easy as in MapReduce */
        /* Reading the last application run from the history server */
        
        /* First, we need to wait for the server to get the last job */
        /* Ten seconds to make sure. Seven proved not to be enough. */
        System.out.println("\n\n**************************************************************************************\n"
        		+ "\tWaitin' for the History Server to get the last job.\n"
        		+ "\tBe patient dude. This is a test, we're not goin' for performance here."
        		+ "\n**************************************************************************************\n\n");
        Thread.sleep(10000);
        
        URL histServer = new URL("http://localhost:18080/api/v1/applications");
        /* The readin' buffer */
		BufferedReader reader = new BufferedReader(new InputStreamReader(histServer.openStream()));
		String readBuff = new String();
		
		/* First line to be ignored (but we got to read it) */
		readBuff = reader.readLine();
		/* This is our line */
		readBuff = reader.readLine();
		String appId = readBuff.substring(10, 33);
		System.out.println("\n\nThe app ID is: " + appId + "\n\n");
		
		
		try
		{
			/* The history server address */
			URL histServer2 = new URL("http://localhost:18080/api/v1/applications/" + appId + "/jobs");
			//URL histServer2 = new URL("http://localhost:18080/history/" + appId + "/stages");
			
			/* The readin' buffer */
			BufferedReader reader2 = new BufferedReader(new InputStreamReader(histServer2.openStream()));
			String readBuff2 = new String();
			
			/* The writin' buffer (to XML file out.xml) */			
			BufferedWriter bwf = new BufferedWriter(new FileWriter("/home/goiko/datos/out" + e + ".txt"));
			
			while((readBuff2 = reader2.readLine()) != null)
			{
				bwf.write(readBuff2);
				bwf.write("\n");
			} // End while loop
			
			bwf.close();
			
		} // End try
		
		catch (Exception excpt)
		{
			excpt.printStackTrace();
		} // End catch
		
		/* Now going after the metrics file */
		URL metricsServer = new URL("http://localhost:18080/api/v1/applications/" + appId + "/jobs");
		InputStream metricsStream = metricsServer.openStream();      
        
        JsonReader metricsRdr = Json.createReader(metricsStream);
        JsonArray completeObject = metricsRdr.readArray();
        JsonObject submitTime = completeObject.getJsonObject(0);
        JsonValue submitted = submitTime.get("submissionTime");
        JsonValue finished = submitTime.get("completionTime");
        
        System.out.println("\nCompletion Time: " + finished.toString());
        System.out.println("Submission time: " + submitted.toString() + "\n");
        
                
        
        
        
        
		} // END OF EXECUTIONS LOOP
        
	} // End method main()

} // End class SparkTraffivGen
